class ConfigError(Exception):
    ...


class ComponentNotFound(ConfigError):
    ...


class KeyNotInConfig(ConfigError):
    ...


class InvalidConfigKey(ConfigError):
    ...


class TypeNotDefined(ConfigError):
    ...


class AppMergeCollisions(ConfigError):
    ...


class InvalidConfigValue(ConfigError):
    ...


class DoesNotSupportBuild(ConfigError):
    ...


class FileTypeNotRecognized(ConfigError):
    ...



import json

from collections.abc import MutableMapping
from pathlib import Path
from typing import Dict
from typing import Text

import toml
import yaml

from config_injector.config import SupportsFill
from config_injector.config import fill
from config_injector.exc import FileTypeNotRecognized


class Injector(MutableMapping):
    def __init__(self, context: Dict = None):
        self.context = {} if context is None else context

    def __getitem__(self, k: Text):
        v = self.context[k]
        if hasattr(v, "items"):
            return self.__class__(v)
        else:
            return v

    def __iter__(self):
        return iter(self.context)

    def __setitem__(self, k, v):
        self.context[k] = v

    def __delitem__(self, k):
        self.context.pop(k)

    def __len__(self):
        return len(self.context)

    def clear(self):
        self.context = {}

    def load(self, context: Dict):
        self.context.update(context)

    def load_file(self, file: Path):
        if file.name.lower().endswith(".json"):
            self._load_json_file(file)
        elif file.name.lower().endswith(".toml"):
            self._load_toml_file(file)
        elif file.name.lower().endswith(".yaml") or file.name.lower().endswith(".yml"):
            self._load_yaml_file(file)
        else:
            raise FileTypeNotRecognized(
                f"Unable to determine file type for {file.name}"
            )

    def _load_json_file(self, file: Path):
        with file.open() as f:
            self.load(json.load(f))

    def _load_toml_file(self, file: Path):
        with file.open() as f:
            self.load(toml.load(f))

    def _load_yaml_file(self, file: Path):
        with file.open() as f:
            self.load(yaml.load(f, Loader=yaml.SafeLoader))

    def instantiate(self, config: SupportsFill):
        return fill(config, self.context)



from abc import ABC
from typing import Any
from typing import Callable
from typing import Dict
from typing import List
from typing import SupportsFloat
from typing import SupportsInt
from typing import Text
from typing import Tuple
from typing import Union

from config_injector.exc import DoesNotSupportBuild
from config_injector.exc import InvalidConfigValue
from config_injector.exc import KeyNotInConfig
from config_injector.exc import TypeNotDefined
from config_injector.utils import get_type


JsonTypes = Union[Dict, List, bool, SupportsInt, SupportsFloat, Text]
ComponentCallable = Union[Any, Tuple[Any]]


class SupportsFill(ABC):
    def __fill__(self, **kwargs: Dict):
        ...


def fill(f: SupportsFill, context: Dict) -> Any:
    try:
        return f.__fill__(**context)
    except AttributeError as e:
        if not hasattr(f, "__build__"):
            raise DoesNotSupportBuild(f"{f} does not support build.", e)
        else:
            raise e


class Config(SupportsFill):
    def __init__(self, callback: Callable, **arg_callables: ComponentCallable):
        """
        A configurable component containing hints for json parsing.

        :param callback: The function to call. The function should return an object.
        :param arg_callables: The callables for the arguments to the callback with matching
           argument names. These callables can be overloaded by providing a tuple.
        """
        self.callback = callback
        self.arg_callables = arg_callables

    def __call__(self, *args, **kwargs) -> Any:
        return self.callback(*args, **kwargs)

    @property
    def __name__(self):
        if hasattr(self.callback, "__name__"):
            return self.callback.__name__
        else:
            raise AttributeError(f"{self.callback} has no attribute {__name__}")

    def get_arg_type(self, arg_name, arg):
        try:
            _arg_tp = self.arg_callables[arg_name]
        except KeyError as e:
            raise e
        try:
            type_map = {get_type(x): x for x in _arg_tp}
        except TypeError:
            type_map = None
        if type_map:
            # Handle "oneOf" types defined with a type property.
            try:
                type_name = arg.pop("type")
            except KeyError:
                raise KeyNotInConfig(
                    f"Config key type not found in configuration arguments for {arg_name}"
                )
            except TypeError:
                raise InvalidConfigValue(f"Config value for {arg_name} is invalid.")
            try:
                arg_tp = type_map[type_name]
            except KeyError:
                raise TypeNotDefined(
                    f"Configuration type {type_name} not defined for {arg_name}"
                )
        else:
            arg_tp = _arg_tp
        return arg_tp

    def __fill__(self, **kwargs: JsonTypes) -> Any:
        """
        Cast data from parsed json prior to calling the callback.

        :param kwargs: Key word arguments consisting of only json types.
        :return:
        """
        kwargs_cast = {}
        for arg_name, arg in kwargs.items():
            arg_tp = self.get_arg_type(arg_name, arg)
            if arg_name in self.arg_callables:
                if hasattr(arg, "items") and hasattr(arg_tp, "__fill__"):
                    arg_cast = fill(arg_tp, arg)
                else:
                    arg_cast = arg_tp(arg)
            else:
                arg_cast = arg
            kwargs_cast[arg_name] = arg_cast
        return self(**kwargs_cast)


def config(**arg_callables: ComponentCallable) -> Callable[[], Config]:
    """
    Decorator for functions with hints for json decoding.

    :param key: The key to use for the configuration.
    :param kwargs: The type for each argument in the function f.
    :return: Wrapper
    """

    def wrapper(f) -> Config:
        component = Config(f, **arg_callables)
        return component

    return wrapper



try:
    from importlib.metadata import version as get_version
except ImportError:
    from importlib_metadata import version as get_version

from config_injector.config import config
from config_injector.injector import Injector


__version__ = get_version(__package__)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['config_injector']

package_data = \
{'': ['*']}

install_requires = \
['PyYAML>=5.3.1,<6.0.0', 'pyaml>=20.4.0,<21.0.0', 'toml>=0.10.2,<0.11.0']

extras_require = \
{':python_version < "3.8"': ['importlib-metadata>=0.12,<=4']}

setup_kwargs = {
    'name': 'config-injector',
    'version': '0.2.1',
    'description': 'Simple dependency injection framework for python for easy and logical app configuration.',
    'long_description': '# config-injector\nConfig-injector is a very simple framework which aims to do only two things: (1) define configurable functions and (2) inject configuration data into those functions at runtime.\n\n## Installation\nInstall with pip.\n```bash\npip install config-injector\n```\n\n## Getting Started\nAnnotate any callable as a configurable function using `@config`. Note that the `@config` decorator requires that you provide callable functions for each argument. These callable functions should return the expected type. The object is to break all arguments down to fundamental types: string, integer, float or dictionary.\n\n```python\nfrom collections import namedtuple\nfrom typing import Text, Dict, SupportsInt\nfrom pathlib import Path\n\nfrom config_injector import config, Injector\n\n\nMockThing0 = namedtuple("MockThing0", ["arg_1", "arg_2", "arg_3", "arg_4"])\n\n@config(arg_1=str, arg_2=str, arg_3=str, arg_4=str)\ndef mock_thing_0(arg_1: Text, arg_2: Text, arg_3: Text, arg_4: Text):\n    return MockThing0(arg_1, arg_2, arg_3, arg_4)\n\n\n@config(arg_5=int, arg_6=int, arg_7=int, arg_8=int)\ndef mock_thing_1(arg_5, arg_6, arg_7, arg_8):\n    return {"key_a": arg_5, "key_b": arg_6, "key_c": arg_7, "key_d": arg_8}\n\n@config(t0=mock_thing_0, t1=mock_thing_1, arg_9=str)\ndef mock_things(t0: MockThing0, t1: Dict[SupportsInt], arg_9: Text):\n    return (t0, t1, arg_9)\n\ndef get_things(config_file=Path("config.json")):\n    injector = Injector()\n    injector.load_file(config_file)\n    return injector["things"].instantiate(mock_things)\n```\n\nNow that the configurable functions are annotated, we can write a configuration for them.\n\n```json\n{\n  "things": {\n    "t0": {"arg_1": "a", "arg_2": "b", "arg_3": "c", "arg_4": "d"},\n    "t1": {"arg_5": 1, "arg_6": 2, "arg_7": 3, "arg_8": 4},\n    "arg_9": "e"\n  }\n}\n```\n\nThis configuration file can be loaded in the runtime portion of our implementation using `get_things()` to instantiate the configured objects created by our functions.',
    'author': 'DustinMoriarty',
    'author_email': 'dustin.moriarty@protonmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/DustinMoriarty/config-injector',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)
